# @(#) $Id: bsd.arduino.mk 73 2020-09-13 17:25:38Z leres $ (XSE)
#
#  bsd.arduino.mk -- build Arduino sketches with make(1)
# 
#
# Copyright (c) 2017, 2018, 2019, 2020
#	Craig Leres
# 	All rights reserved
#
# Author: Craig Leres <leres@xse.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#

#
# +++ variables +++
#
# These must be set:
#
# TARGET		Name of the sketch
# ARDUINO_BOARD		Board name
# ARDUINO_MK_DIR	Path to arduino-bsd-mk installation
#
# It's frequently useful to set these:
#
# SRCS			List of .ino, .cpp and .c sources
# ARDUINO_LIBS		List of Arduino libraries
#
# Programming options:
#
# boards.txt is parsed to populate the defaults for avrdude options
# Thse can be overiden in your Makefile or Makefile.inc
#
# AVRDUDE_PROGRAMMER	avrdude programmer id string
#
# When the programmer is "arduino" AVRDUDE_PORT must using one of these:
#
#    - In your shell environment before running "make flash"
#    - With the make command line, e.g. "make flash AVRDUDE_PORT=/dev/ttyU3"
#    - In Makefile.inc, e.g. "AVRDUDE_PORT=/dev/ttyU3"
#
# When the programmer is one that uses USB, AVRDUDE_PORT can
# optionally specify the usb port. See the -P option in the avrdude(1)
# man page for more details.
#
# AVRDUDE_PORT		avrdude programmer port
#
# AVRDUDE_SPEED		avrdude programmer serial speed (programmer: arduino)
#
# These can be set:
#
# ARDUINO_DIR		Top level of arduino installation
# AVR_LIBC_DIR		Top level of arv-libc installation
#
# F_CPU			Speed of processor
# ARDUINO_CFLAGS	Additional CFLAGS for sketch modules
# ARDUINO_CXXFLAGS	Additional CXXFLAGS for sketch modules
# HFILES		Includes files (used for tags)
#
# Debugging:
#
# WITH_CPPCHECK_XML	Run cppcheck with XML output enabled
# WITH_CPPCHECK_CHECKCONFIG Run cppcheck with --check-config
#

.if !defined(TARGET)
.error Please define TARGET
.endif

.if !defined(ARDUINO_BOARD)
.error Please define ARDUINO_BOARD
.endif

.if exists(${.CURDIR}/Makefile.inc)
.include "${.CURDIR}/Makefile.inc"
.endif

PROG=			${TARGET}.elf

all:			${TARGET}.hex

.if !defined(SRCS)
.if exists(${.CURDIR}/${TARGET}.ino)
SRCS=			${TARGET}.ino
.else
SRCS=			${TARGET}.cpp
.endif
.endif

# No man page by default
MAN?=

# No debugging binaries by default
MK_DEBUG_FILES?=	no

PREFIX?=		/usr/local
ARDUINO_PREFIX?=	${PREFIX}

ARDUINO_DIR?=		${ARDUINO_PREFIX}/arduino
ARDUINO_LIBRARIES_DIR?=	${ARDUINO_DIR}/libraries

ARDUINO_SRC_DIR?=	${ARDUINO_DIR}/hardware/arduino/cores/arduino
ARDUINO_VAR_DIR?=	${ARDUINO_DIR}/hardware/arduino/variants/${ARDUINO_VARIANT}

AVR_LIBC_DIR?=		${PREFIX}/avr
AVR_PREFIX?=		/usr/local

LIBDIR?=		${AVR_LIBC_DIR}/lib
LIBDIR_BASE?=		${AVR_LIBC_DIR}/lib

CC=			avr-gcc
CXX=			avr-g++
AR=			avr-ar
OBJCOPY=		avr-objcopy
SIZE=			avr-size

AVRDUDE?=		avrdude
CPPCHECK?=		cppcheck
ARDUINO_BOARDS?=	${ARDUINO_MK_DIR}/scripts/arduino-boards

INCLS?=			-I. -I${.CURDIR}

COPT?=			-Os
CWFLAGS?=		-fno-exceptions -ffunction-sections -fdata-sections
DEFS?=			-DF_CPU=${F_CPU} -DARDUINO=${ARDUINO_VERSION}

#
# Things we determine at run time
#
ARDUINO_VERSION?=	${:!${ARDUINO_MK_DIR}/scripts/arduino-version ${ARDUINO_DIR}/lib/version.txt!}

ARDUINO_CORE?=		${:!${ARDUINO_BOARDS} ${ARDUINO_BOARD} build.core!}
ARDUINO_VARIANT?=	${:!${ARDUINO_BOARDS} ${ARDUINO_BOARD} build.variant!}
F_CPU?=			${:!${ARDUINO_BOARDS} ${ARDUINO_BOARD} build.f_cpu!}
MCU?=			${:!${ARDUINO_BOARDS} ${ARDUINO_BOARD} build.mcu!}

USB_PID?=		${:!${ARDUINO_BOARDS} -e ${ARDUINO_BOARD} build.pid!}
.if !empty(USB_PID)
DEFS+=			-DUSB_PID=${USB_PID}
.endif

USB_VID?=		${:!${ARDUINO_BOARDS} -e ${ARDUINO_BOARD} build.vid!}
.if !empty(USB_VID)
DEFS+=			-DUSB_VID=${USB_VID}
.endif

GCC_VER=		${:!${ARDUINO_MK_DIR}/scripts/gcc-version!}

CFLAGS=			-mmcu=${MCU} -g -std=gnu99 ${COPT} ${CWFLAGS} ${DEFS} ${ARDUINO_CFLAGS} ${INCLS}
CXXFLAGS=		-mmcu=${MCU} -g -std=c++11 ${COPT} ${CWFLAGS} ${DEFS} ${ARDUINO_CXXFLAGS} ${INCLS}

LDFLAGS=		-mmcu=${MCU} -Wl,--gc-sections ${COPT}

# Build a library arduino modules
ARDUINO_LIB=		libarduino.a

.PATH: ${ARDUINO_SRC_DIR}
INCLS+=			-I${ARDUINO_SRC_DIR} \
			-I${ARDUINO_VAR_DIR}

LIB_SRCS=		${:!ls ${ARDUINO_SRC_DIR} | egrep '\.(c|cpp)\$' || true!:T}

LIB_OBJS=		${LIB_SRCS:R:S/$/.o/g}
CLEANFILES+=		${LIB_OBJS}

${TARGET}.elf:		${ARDUINO_LIB}

LDADD+=			${ARDUINO_LIB}
CLEANFILES+=		${ARDUINO_LIB}

# Add in the sources and directories for any Arduino libraries
.for _LIB in ${ARDUINO_LIBS}

# Try local library first
.if exists(${.CURDIR}/${_LIB})
.for _DN in ${.CURDIR}/${_LIB}
.PATH.cpp:		${_DN}
.PATH.c:		${_DN}
LIB_SRCS+=		${:!ls ${_DN} | egrep '\.(c|cpp)\$' || true!:T}
INCLS+=			-I${_DN}

# Add the utility subdirectory if it exists
.if exists(${_DN}/utility)
.PATH.cpp:		${_DN}/utility
.PATH.c:		${_DN}/utility
LIB_SRCS+=		${:!ls ${_DN}/utility | egrep '\.(c|cpp)\$' || true!:T}
INCLS+=			-I${_DN}/utility
.endif
.endfor

# Try arduino libraries version
.elif exists(${ARDUINO_LIBRARIES_DIR}/${_LIB})
.for _DN in ${ARDUINO_LIBRARIES_DIR}/${_LIB}
.PATH.cpp:		${_DN}
.PATH.c:		${_DN}
LIB_SRCS+=		${:!ls ${_DN} | egrep '\.(c|cpp)\$' || true!:T}
INCLS+=			-I${_DN}

# Add the utility subdirectory if it exists
.if exists(${_DN}/utility)
.PATH.cpp:		${_DN}/utility
.PATH.c:		${_DN}/utility
LIB_SRCS+=		${:!ls ${_DN}/utility | egrep '\.(c|cpp)\$' || true!:T}
INCLS+=			-I${_DN}/utility
.endif
.endfor

.else
.error ${_LIB} doesn't exist
.endif

.endfor

# Build Arduino modules with -w (inhibit all warnings messages)
# Remove objs after building .a
${ARDUINO_LIB}: ${LIB_SRCS}
	${MAKE} -C ${.CURDIR} CFLAGS="-w ${CFLAGS}" \
	    CXXFLAGS="-w ${CXXFLAGS}" ${LIB_OBJS}
	rm -f ${.TARGET}
	${AR} ${ARFLAGS} ${.TARGET} ${LIB_OBJS}
	rm -f ${LIB_OBJS}

# elf to hex transformation
.SUFFIXES: .elf .hex

CLEANFILES+=		${TARGET}.hex

.elf.hex:
	${OBJCOPY} -O ihex -R .eeprom ${.IMPSRC} ${.TARGET}

# .ino cleanup
.SUFFIXES: .ino

# Cleanup .cpp when .ino exists
.for _INOSRC in ${SRCS:M*.ino:N*/*}
.for _CPPSRC in ${_INOSRC:R}.cpp
# cleanup
CLEANFILES+=	${.OBJDIR}/${_CPPSRC}
.endfor
.endfor

# Disallow .ino when .cpp exists
.for _CPPSRC in ${SRCS:M*.cpp:N*/*}
.for _INOSRC in ${_CPPSRC:R}.ino
.if exists(${.CURDIR}/${_INOSRC}) && exists(${.CURDIR}/${_CPPSRC})
.error Both ${_CPPSRC} and ${_INOSRC} exist
.endif
.endfor
.endfor

# ino to cpp transformation
.ino.cpp:
	${ECHO} rm -f ${.IMPSRC:T:R}.cpp
	${ECHO} '#include <Arduino.h>' > ${.IMPSRC:T:R}.cpp
	${ECHO} '#line 1 "${.IMPSRC:T:R}.ino"' >> ${.IMPSRC:T:R}.cpp
	cat ${.IMPSRC} >> ${.IMPSRC:T:R}.cpp

#
# assembler
#
.SUFFIXES: .s

.cpp.s:
	${CXX} ${CXXFLAGS} -S ${.IMPSRC} -o ${.TARGET}

#
# depend: use generated .cpp instead of .ino sources
#
depend: ${SRCS:N*.ino} ${SRCS:M*.ino:S/.ino/.cpp/:T} ${HFILES}
	@env CC=${CXX} mkdep ${CXXFLAGS} ${.ALLSRC}

#
# tags
#
CTAGS?=		ctags
CTAGSFLAGS?=	-wtd
.if defined(SRCS) && !defined(NO_TAGS)
tags: ${SRCS} ${HFILES}
	@${CTAGS} ${CTAGSFLAGS} -f /dev/stdout \
	    ${.ALLSRC:N*.tags} | sed "s;${.CURDIR}/;;" > ${.TARGET}
	@test -h ${.CURDIR}/obj -a ! -h ${.CURDIR}/tags && \
	    ln -s obj/tags ${.CURDIR}/tags || true
.endif

#
# defines: dump out all including the board specifc ones
#
defines:
	${CC} -mmcu=${MCU} -dM -E - < /dev/null | sort

#
# lint
#

LINT_INCLS= \
	-I${PREFIX}/lib/gcc/avr/${GCC_VER}/include \
	-I${PREFIX}/lib/gcc/avr/${GCC_VER}/include-fixed \
	-I${AVR_LIBC_DIR}/include \
	${INCLS}

CPPCHECK_FLAGS= \
		-q \
		--enable=all \
		-Dlint \
		-DTCNT0=1 \
		-DF_CPU=${F_CPU} \
		-DARDUINO=${ARDUINO_VERSION}

AVR_DEF?=		__AVR_${MCU:tu:S/MEGA/mega/}__
CPPCHECK_FLAGS+=	-D${AVR_DEF}=1

.if defined(WITH_CPPCHECK_XML)
CPPCHECK_FLAGS+=	--xml --xml-version=2
.endif

.if defined(WITH_CPPCHECK_CHECKCONFIG)
CPPCHECK_FLAGS+=	--check-config
.endif

# The scope of the variable '%s' can be reduced
CPPCHECK_FLAGS+= --suppress=variableScope

# Noise reduction
CPPCHECK_FLAGS+= --suppress=noExplicitConstructor:${ARDUINO_SRC_DIR}/IPAddress.h
CPPCHECK_FLAGS+= --suppress=noExplicitConstructor:${ARDUINO_SRC_DIR}/WString.h
CPPCHECK_FLAGS+= --suppress=preprocessorErrorDirective:${ARDUINO_SRC_DIR}/WInterrupts.c
CPPCHECK_FLAGS+= --suppress=preprocessorErrorDirective:${ARDUINO_SRC_DIR}/wiring.c
CPPCHECK_FLAGS+= --suppress=uninitMemberVar:${ARDUINO_SRC_DIR}/HardwareSerial.cpp
CPPCHECK_FLAGS+= --suppress=uninitMemberVar:${ARDUINO_SRC_DIR}/Stream.h
CPPCHECK_FLAGS+= --suppress=unsignedLessThanZero:${ARDUINO_SRC_DIR}/WString.cpp
CPPCHECK_FLAGS+= --suppress=unusedFunction:${ARDUINO_SRC_DIR}/Stream.cpp
CPPCHECK_FLAGS+= --suppress=unusedFunction:${ARDUINO_SRC_DIR}/Tone.cpp
CPPCHECK_FLAGS+= --suppress=unusedFunction:${ARDUINO_SRC_DIR}/WMath.cpp
CPPCHECK_FLAGS+= --suppress=unusedFunction:${ARDUINO_SRC_DIR}/WString.cpp
CPPCHECK_FLAGS+= --suppress=unusedFunction:${ARDUINO_SRC_DIR}/main.cpp
CPPCHECK_FLAGS+= --suppress=unusedFunction:${ARDUINO_SRC_DIR}/new.cpp
CPPCHECK_FLAGS+= --suppress=unusedFunction:${ARDUINO_SRC_DIR}/wiring_analog.c
CPPCHECK_FLAGS+= --suppress=unusedFunction:${ARDUINO_SRC_DIR}/wiring_pulse.c
CPPCHECK_FLAGS+= --suppress=unusedFunction:${ARDUINO_SRC_DIR}/wiring_shift.c

lint: ${SRCS} ${LIB_SRCS}
.for _INCL in ${LIT_INCLS}
.if !exists(${_INCL})
.error ${_INCL} doesn't exist
.endfi
.endfor
	cd ${.OBJDIR} && ${CPPCHECK} ${CPPCHECK_FLAGS} ${LINT_INCLS} ${.ALLSRC}

#
# Programming
#
AVRDUDE_PROGRAMMER?=	${:!${ARDUINO_BOARDS} ${ARDUINO_BOARD} upload.protocol!}
.if ${AVRDUDE_PROGRAMMER} == "arduino"
AVRDUDE_SPEED?=	${:!${ARDUINO_BOARDS} ${ARDUINO_BOARD} upload.speed!}
AVRDUDE_FLAGS?=		-b ${AVRDUDE_SPEED} -P ${AVRDUDE_PORT}
.else
.if defined(AVRDUDE_PORT)
AVRDUDE_FLAGS?=		-P ${AVRDUDE_PORT}
.endif
.endif

install flash: ${TARGET}.hex
.if ${AVRDUDE_PROGRAMMER} == "arduino" && !defined(AVRDUDE_PORT)
	@echo 'Define AVRDUDE_PORT when AVRDUDE_PROGRAMMER is "arduino"'
	@false
.else
	${AVRDUDE} \
	    -V \
	    -p ${MCU} \
	    -c ${AVRDUDE_PROGRAMMER} \
	    ${AVRDUDE_FLAGS} \
	    -U flash:w:${.OBJDIR}/${TARGET}.hex:i
.endif

# Report the fuse settings
report:
	@${ECHO} ">> Using ${AVRDUDE} to query ${MCU} via ${AVRDUDE_PROGRAMMER}"
	@${AVRDUDE} \
	    -V \
	    -p ${MCU} \
	    -c ${AVRDUDE_PROGRAMMER} \
	    -qq \
	    -U lfuse:r:/dev/stdout:h \
	    -U hfuse:r:/dev/stdout:h \
	    -U efuse:r:/dev/stdout:h \
	    -U lock:r:/dev/stdout:h | awk -v part="${MCU}" ' \
		BEGIN { \
			n = 1; \
			url = "http://eleccelerator.com/fusecalc/fusecalc.php"; \
		} \
		{ \
			input[n] = $$0; \
			++n; \
		} \
		END { \
			if (n < 4) { \
				printf "Missing output from avrdude (%d)\n", n; \
				exit 1; \
			} \
			i = 1; \
			lfuse = 0 + input[i++]; \
			hfuse = 0 + input[i++]; \
			efuse = 0 + input[i++]; \
			lock = 0 + input[i++]; \
			printf "lfuse: 0x%02X\n", lfuse; \
			printf "hfuse: 0x%02X\n", hfuse; \
			printf "efuse: 0x%02X\n", efuse; \
			printf " lock: 0x%02X\n", lock; \
			printf "  url: %s?chip=%s&LOW=%x&HIGH=%x&EXTENDED=%x&LOCKBIT=%x\n", url, part, lfuse, hfuse, efuse, lock; \
		} \
		'

size:
	${SIZE} -B ${TARGET}.elf

# Add these last
LDADD+=			-lc -lm

# XXX prevent ssp junk from kicking in
MACHINE_CPUARCH=	arm
.include <bsd.prog.mk>
